class AuthException(Exception):
    pass


class ResponseException(Exception):
    pass


class ParentNotFoundException(Exception):
    pass


class EmptyException(Exception):
    pass


class ResourceNotFound(Exception):
    pass


class NotAnInstanceException(Exception):
    pass
